import os

from codercore.db import get_connection_url


class EnvSettings:
    @classmethod
    @property
    def database_connection_url(cls) -> str:
        return get_connection_url(
            "postgresql+asyncpg",
            os.environ["POSTGRES_USER"],
            os.environ["POSTGRES_PASSWORD"],
            os.environ["POSTGRES_HOST"],
            os.environ["POSTGRES_DB"],
        )

    @classmethod
    @property
    def redis(cls) -> dict:
        return {
            "host": os.environ["REDIS_HOST"],
            "port": int(os.environ["REDIS_PORT"]),
        }



from functools import cache
from typing import Type, Optional

from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import sessionmaker as sessionmaker_, Session as Session_
from sqlalchemy.pool import Pool

from codercore.db.models import Base


@cache
def Session(connection_url: str) -> Session_:  # noqa
    return sessionmaker(connection_url)()


@cache
def sessionmaker(
    connection_url: str,
    *args,
    poolclass: Optional[Type[Pool]] = None,
    **kwargs,
) -> sessionmaker_:
    engine = create_async_engine(connection_url, poolclass=poolclass)
    Base.metadata.bind = engine
    return sessionmaker_(engine, *args, class_=AsyncSession, **kwargs)


def get_connection_url(
    driver: str, user: str, password: str, host: str, database: str
) -> str:
    return f"{driver}://{user}:{password}@{host}/{database}"



from __future__ import annotations

from functools import cache
from typing import Any, Awaitable, Optional, Union

from redis.asyncio import StrictRedis


@cache
def connection(
    host: str,
    port: int = 6379,
    db: int = 0,
    password: Optional[str] = None,
    default_ttl_in_seconds: Optional[int] = None,
) -> Redis:
    return Redis(
        host=host,
        port=port,
        db=db,
        password=password,
        default_ttl_in_seconds=default_ttl_in_seconds,
    )


class Redis(StrictRedis):
    default_ttl_in_seconds: Optional[int]

    def __init__(
        self,
        *args,
        default_ttl_in_seconds: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.default_ttl_in_seconds = default_ttl_in_seconds

    async def set(
        self,
        *args,
        ex: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable, Any]:
        if not ex:
            ex = self.default_ttl_in_seconds
        await super().set(*args, ex=ex, **kwargs)



from collections.abc import AsyncIterator

from pytest import FixtureRequest
from sqlalchemy import MetaData
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import sessionmaker as sessionmaker_
from sqlalchemy.pool import NullPool
from sqlalchemy_utils import database_exists, create_database, drop_database

from codercore.db import get_connection_url, sessionmaker
from codercore.db.models import Base
from codercore.lib.redis import connection, Redis
from codercore.lib.settings import EnvSettings


def connection_settings(
    user: str,
    password: str,
    host: str,
    database: str,
) -> dict[str, str]:
    return {
        "user": user,
        "password": password,
        "host": host,
        "database": database,
    }


def sync_db_connection_url(connection_settings: dict[str, str]) -> str:
    return get_connection_url("postgresql", **connection_settings)


def async_db_connection_url(connection_settings: dict[str, str]) -> str:
    return get_connection_url("postgresql+asyncpg", **connection_settings)


def DBSession(  # noqa
    sync_db_connection_url: str,
    async_db_connection_url: str,
    *args,
    **kwargs,
) -> sessionmaker_:
    if not database_exists(sync_db_connection_url):
        create_database(sync_db_connection_url)
    return sessionmaker(async_db_connection_url, *args, poolclass=NullPool, **kwargs)


async def db_session(
    DBSession: sessionmaker_,  # noqa
    metadata: MetaData = Base.metadata,
) -> AsyncIterator[AsyncSession]:
    async with DBSession() as session:
        try:
            async with session.bind.begin() as conn:
                await conn.run_sync(metadata.create_all)
            yield session
        finally:
            async with session.bind.begin() as conn:
                await conn.run_sync(metadata.drop_all)


def clean_up_for_worker(request: FixtureRequest, sync_db_connection_url: str) -> None:
    def cleanup():
        if not database_exists(sync_db_connection_url):
            return

        drop_database(sync_db_connection_url)

    request.addfinalizer(cleanup)


async def redis_connection(worker_id: str) -> AsyncIterator[Redis]:
    redis = connection.__wrapped__(db=int(worker_id[2:]), **EnvSettings.redis)
    yield redis
    await redis.flushdb()
    await redis.close()



from sqlalchemy.orm import as_declarative


@as_declarative()
class Base:
    def set_fields(self, **kwargs) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __hash__(self) -> int:
        raise NotImplementedError

    def __eq__(self, other: any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return hash(self) == hash(other)
